/* eslint-env mocha */
/* eslint-disable no-unused-expressions */
const chai = require('chai')
chai.use(require('chai-as-promised'))
const expect = chai.expect
const nock = require('nock')
const mock = require('mock-fs')
const stdMocksWrapper = require('../support/std-mocks-wrapper')

const Node = require('../../lib/node')

describe('Node', function () {
  before(function () {
    nock('https://www.example.com')
      .persist()
      .get('/test.gitignore')
      .reply(200, 'some web contents\n')

    nock('https://www.example.com')
      .persist()
      .get('/test2.getignore')
      .reply(200, '#=https://www.example.com/test.gitignore\n')

    nock('https://www.example.com')
      .persist()
      .get('/delayed.gitignore')
      .delay(50)
      .reply(200, 'some delayed web contents\n')

    nock('https://www.example.com')
      .persist()
      .get('/not-found.gitignore')
      .reply(404)

    nock('https://raw.githubusercontent.com')
      .persist()
      .get('/github/gitignore/master/test.gitignore')
      .reply(200, 'some github/gitignore contents\n')

    nock('https://raw.githubusercontent.com')
      .persist()
      .get('/github/gitignore/master/test-not-found.gitignore')
      .reply(400)
  })

  after(function () {
    nock.restore()
  })

  beforeEach(function () {
    mock({
      'test.gitignore': 'a/non-getignore.rule\n',
      'test/test.gitignore': 'a/**sub/non-getignore.rule\n',
      'test/up.gitignore': '#=file://../test.gitignore\n',
      '.getignore': 'a/non-getignore.rule\n#=file://test/up.gitignore\n#==test-not-found\nblub\n#=file://test-not-there.gitignore\n#=https://www.example.com/test.gitignore\n#==test',
      'delaying.getignore': '#=https://www.example.com/delayed.gitignore\n#=https://www.example.com/test.gitignore\n'
    })
  })

  afterEach(function () {
    mock.restore()
  })

  describe('conformURL', function () {
    it('should treat URLs beginning with = as GitHub/gitignore URLs', function () {
      expect(Node.conformURL('=test')).to.equal('https://raw.githubusercontent.com/github/gitignore/master/test.gitignore')
    })

    it('should keep URLs beginning with =', function () {
      expect(Node.conformURL('test')).to.equal('test')
    })
  })

  describe('getContents', function () {
    it('should get file contents', function () {
      return expect(Node.getContents('file://test.gitignore')).to.become('## getignore: test.gitignore ##\na/non-getignore.rule\n')
    })

    it('should get file contents in a sub directiory', function () {
      return expect(Node.getContents('file://test/test.gitignore')).to.become('## getignore: test/test.gitignore ##\na/**sub/non-getignore.rule\n')
    })

    it('should get file contents and skip header', function () {
      return expect(Node.getContents('file://test.gitignore', { skipHeader: true })).to.become('a/non-getignore.rule\n')
    })

    it('should get file contents relative to parent', function () {
      return Promise.all([
        expect(Node.getContents('file://test.gitignore', { parent: 'test' })).to.become('## getignore: test/test.gitignore ##\na/**sub/non-getignore.rule\n'),
        expect(Node.getContents('file://../test.gitignore', { parent: 'test' })).to.become('## getignore: test.gitignore ##\na/non-getignore.rule\n'),
        expect(Node.getContents('file://../test.gitignore', { parent: 'test-not-there' })).to.become('## getignore: test.gitignore ##\na/non-getignore.rule\n')
      ])
    })

    it('should get file contents and notify fs:// is deprecated', function () {
      return stdMocksWrapper(() => Node.getContents('fs://test.gitignore')).then(_ => {
        expect(_.value).to.equal('## getignore: test.gitignore ##\na/non-getignore.rule\n')
        expect(_.stdout).to.eql(['Warning: "fs://" is deprecated - Use "file://" instead.\n'])
        expect(_.stderr).to.be.empty
      })
    })

    it('should gracefully notify when file does not exist', function () {
      return stdMocksWrapper(() => Node.getContents('file://test-not-there.gitignore')).then(_ => {
        expect(_.value).to.equal('## getignore: error: test-not-there.gitignore ##\n')
        expect(_.stdout).to.eql(['Error:   test-not-there.gitignore\n'])
        expect(_.stderr).to.be.empty
      })
    })

    it('should gracefully notify on unsupported URLs', function () {
      return stdMocksWrapper(() => Node.getContents('blub://nope')).then(_ => {
        expect(_.value).to.equal('## getignore: error: blub://nope ##\n')
        expect(_.stdout).to.eql(['Error:   blub://nope\n'])
        expect(_.stderr).to.be.empty
      })
    })

    it('should get web contents', function () {
      return stdMocksWrapper(() => Node.getContents('https://www.example.com/test.gitignore')).then(_ => {
        expect(_.value).to.equal('## getignore: https://www.example.com/test.gitignore ##\nsome web contents\n')
        expect(_.stdout).to.eql(['Success: https://www.example.com/test.gitignore\n'])
        expect(_.stderr).to.be.empty
      })
    })

    it('should get web contents and skip header', function () {
      return stdMocksWrapper(() => Node.getContents('https://www.example.com/test.gitignore', { skipHeader: true })).then(_ => {
        expect(_.value).to.equal('some web contents\n')
        expect(_.stdout).to.eql(['Success: https://www.example.com/test.gitignore\n'])
        expect(_.stderr).to.be.empty
      })
    })

    it('should fail on missing web contents', function () {
      return stdMocksWrapper(() => Node.getContents('https://www.example.com/not-found.gitignore')).then(_ => {
        expect(_.value).to.equal('## getignore: error: https://www.example.com/not-found.gitignore ##\n')
        expect(_.stdout).to.eql(['Error:   https://www.example.com/not-found.gitignore\n'])
        expect(_.stderr).to.be.empty
      })
    })
  })

  describe('evaluate', function () {
    it('should evaluate lines part of getignore syntax mixed with lines not part of getignore syntax and gracefully notify about errors', function () {
      return stdMocksWrapper(() => Node.evaluate('file://.getignore', { skipHeader: true })).then(_ => {
        expect(_.value).to.equal('a/non-getignore.rule\n## getignore: test/up.gitignore ##\n## getignore: test.gitignore ##\na/non-getignore.rule\n\n## getignore: error: https://raw.githubusercontent.com/github/gitignore/master/test-not-found.gitignore ##\n\nblub\n## getignore: error: test-not-there.gitignore ##\n\n## getignore: https://www.example.com/test.gitignore ##\nsome web contents\n\n## getignore: https://raw.githubusercontent.com/github/gitignore/master/test.gitignore ##\nsome github/gitignore contents\n')
        expect(_.stdout).to.be.an('array').with.lengthOf(4).that
          .includes('Error:   test-not-there.gitignore\n')
          .includes('Error:   https://raw.githubusercontent.com/github/gitignore/master/test-not-found.gitignore\n')
          .includes('Success: https://www.example.com/test.gitignore\n')
          .includes('Success: https://raw.githubusercontent.com/github/gitignore/master/test.gitignore\n')
        expect(_.stderr).to.be.empty
      })
    })

    it('should evaluate relative paths upwards (file)', function () {
      return stdMocksWrapper(() => Node.evaluate('file://test/up.gitignore')).then(_ => {
        expect(_.value).to.equal('## getignore: test/up.gitignore ##\n## getignore: test.gitignore ##\na/non-getignore.rule\n')
        expect(_.stdout).to.be.empty
        expect(_.stderr).to.be.empty
      })
    })

    it('should evaluate slow responses and retain ordering', function () {
      this.slow(150)
      return stdMocksWrapper(() => Node.evaluate('file://delaying.getignore')).then(_ => {
        expect(_.value).to.equal('## getignore: delaying.getignore ##\n## getignore: https://www.example.com/delayed.gitignore ##\nsome delayed web contents\n\n## getignore: https://www.example.com/test.gitignore ##\nsome web contents\n')
        expect(_.stdout).to.eql([
          'Success: https://www.example.com/test.gitignore\n',
          'Success: https://www.example.com/delayed.gitignore\n'
        ])
        expect(_.stderr).to.be.empty
      })
    })
  })

  describe('ensureSingleNewLineAtEnd', function () {
    it('should add single new line with no newline at end', function () {
      expect(Node.ensureSingleNewLineAtEnd('1234\nqwer\nasdf')).to.equal('1234\nqwer\nasdf\n')
    })

    it('should keep contents with single newline at end', function () {
      expect(Node.ensureSingleNewLineAtEnd('1234\nqwer\nasdf\n')).to.equal('1234\nqwer\nasdf\n')
    })

    it('should keep contents but squash newlines at end with multiple newlines at end', function () {
      expect(Node.ensureSingleNewLineAtEnd('1234\nqwer\nasdf\n\n')).to.equal('1234\nqwer\nasdf\n')
      expect(Node.ensureSingleNewLineAtEnd('1234\nqwer\nasdf\n\n\n\n')).to.equal('1234\nqwer\nasdf\n')
    })
  })

  describe('getParent', function () {
    it('should extract parent directory name', function () {
      expect(Node.getParent('file://test')).to.equal('file://.')
      expect(Node.getParent('file://test/test')).to.equal('file://test')
      expect(Node.getParent('file:///test')).to.equal('file:///')
      expect(Node.getParent('file:///test/test')).to.equal('file:///test')
      expect(Node.getParent('https://www.example.com/test')).to.equal('https://www.example.com')
      expect(Node.getParent('https://www.example.com/test/test')).to.equal('https://www.example.com/test')
    })
  })
})
